﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using Bosch.VideoSDK;
using Bosch.VideoSDK.Device;
using Bosch.VideoSDK.MediaDatabase;

namespace Bosch.Samples.MediaDbExport
{
    ///<summary>
    /// Connection states
    ///</summary>
    public enum ConnectionState
    {
        Disconnected,
        Connecting,
        Connected,
        Disconnecting,
        Export
    };

    /// <summary>
    /// Main window class.
    /// </summary>
    public partial class MainWindow : Form
    {
        private Bosch.VideoSDK.Device.DeviceConnectorClass deviceConnector = null;
        private Bosch.VideoSDK.Device.DeviceProxy proxy;

        private MediaDatabaseBrowser mediaDb;
        private int tracksCompletelyLoaded;
        private Queue<int> trackToBeLoaded = new Queue<int>();
        
        private MediaFileWriter writer;
        private PlaybackController playback;

        private Track exportTrack;
        private string exportPath;
        private MediaFileFormatEnum exportFormat;
        private DateTime exportTimeStart;
        private DateTime exportTimeEnd;

        /// <summary>
        /// Current connection state
        /// </summary>
        public ConnectionState State { get; set; }

        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();

            // Create the Video SDK's device connector
            deviceConnector = new DeviceConnectorClass();
            deviceConnector.ConnectResult += OnConnectResult;
            deviceConnector.DefaultTimeout = 2000;

            // create media database class
            mediaDb = new MediaDatabaseBrowser();
            mediaDb.Tracks.CollectionChanged += Tracks_CollectionChanged;
            mediaDb.OnStateChanged += mediaDb_OnStateChanged;
            mediaDb.OnProgress += mediaDb_OnProgress;

            // subscribe for log events
            var logTraceListener = new LogTraceListener();
            logTraceListener.LogWrite += OnTraceListenerLogWrite;
            Trace.Listeners.Add(logTraceListener);
        }

        #endregion

        private void MainWindow_Load(object sender, EventArgs e)
        {
            SetConnectionState(ConnectionState.Disconnected, "Disconnected");
            Clear(true);
        }

        #region Connect methods and handlers

        private void buttonConnect_Click(object sender, EventArgs e)
        {
            // connect
            if (State == ConnectionState.Disconnected)
            {
                Clear(true);
                
                string info = string.Format("Connecting to {0}...", textBoxDeviceIP.Text);
                SetConnectionState(ConnectionState.Connecting, info);

                deviceConnector.ConnectAsync(textBoxDeviceIP.Text, "GCA.VIP.DeviceProxy");
            }

            // disconect
            if (State == ConnectionState.Connected)
            {
                SetConnectionState(ConnectionState.Disconnecting, "Disconnecting...");

                proxy.Disconnect();

                SetConnectionState(ConnectionState.Disconnected, "Disconnected");

                Clear(false);
            }
        } 


        private void textBoxDeviceIP_KeyDown(object sender, KeyEventArgs e)
        {
            // perform connect on Enter key in address text box
            if (e.KeyCode == Keys.Enter)
                buttonConnect.PerformClick();
        }

        private void OnConnectResult(ConnectResultEnum connectresult, string url, DeviceProxy deviceProxy)
        {
            proxy = deviceProxy;

            switch (connectresult)
            {
                case ConnectResultEnum.creConnected:
                    break;

                case ConnectResultEnum.creInitialized:  
                    // The connection was successfully established.
                    SetConnectionState(ConnectionState.Connected, "Connected. Searching tracks...");

                    // start searching tracks
                    mediaDb.SearchTracks(proxy.MediaDatabase);
                    break;

                default:
                    proxy.Disconnect();
                    Trace.TraceError("Connect to {0} failed. Error: {1}", url, connectresult);
                    SetConnectionState(ConnectionState.Disconnected, "Connection failed");
                    break;
            }
        }

        private void SetConnectionState(ConnectionState state, string statusDescription)
        {
            Trace.TraceInformation(statusDescription);

            State = state;

            // update Connect button caption
            Action updateButtonAction = (() =>
            {
                switch (State)
                {
                    case ConnectionState.Disconnected:
                        buttonConnect.Text = @"Connect";
                        buttonConnect.Enabled = true;
                        break;
                    case ConnectionState.Connected:
                        buttonConnect.Text = @"Disconnect";
                        buttonConnect.Enabled = true;
                        break;
                    default:
                        buttonConnect.Text = state.ToString();
                        buttonConnect.Enabled = false;
                        break;
                }
            });

            // update status 
            Action updateStatusAction = (() =>
            {
                labelStatusInfo.Text = statusDescription;
            });

            SafeInvoke(buttonConnect, updateButtonAction);
            SafeInvoke(labelStatusInfo, updateStatusAction);
        }
        #endregion

        #region Load media database event handlers
        private void mediaDb_OnProgress(int progress)
        {
            float totalProgress = 0;
            if (mediaDb.State == MediaDatabaseBrowserState.TracksLoading || mediaDb.State == MediaDatabaseBrowserState.TracksLoaded)
            {
                // estimate loading tracks as 10%
                totalProgress = progress / 10;
            }
            else
            {
                // "divide" rest of progress bar (90%) between all tracks
                int tracksCount = mediaDb.Tracks.Count;

                if (tracksCount > 0)
                    totalProgress = 10 + 0.9f * (progress + tracksCompletelyLoaded * 100) / tracksCount;
                else
                    totalProgress = progress;
            }

            SetConnectionState(ConnectionState.Connected, "Connected. Loading tracks info " + (int)totalProgress + "%");
            //Trace.TraceInformation("Total progress: {0}%", (int)totalProgress);
        }

        private void mediaDb_OnStateChanged(MediaDatabaseBrowserState state, string description)
        {
            labelStatusInfo.Text = description;
            labelStatusInfo.Update();

            if (state == MediaDatabaseBrowserState.TracksLoaded)
            {
                // prepare list of tracks for loading records
                trackToBeLoaded.Clear();
                foreach (var track in mediaDb.Tracks)
                {
                    trackToBeLoaded.Enqueue(track.TrackID);
                }

                // start load first
                LoadNextTrackRecords();
            }

            if (state == MediaDatabaseBrowserState.RecordsLoaded)
            {
                tracksCompletelyLoaded++;
                LoadNextTrackRecords();
            }
        }

        private void Tracks_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == NotifyCollectionChangedAction.Add)
            {
                // add track nodes
                foreach(VideoSDK.MediaDatabase.Track track in e.NewItems)
                {
                    string displayName = string.Format("Track #{0} {1}", track.TrackID, track.Name);

                    comboBoxTracks.Items.Add(displayName);
                }
            }
        }

        private void LoadNextTrackRecords()
        {
            // Start loading records for found tracks from queue
            if (trackToBeLoaded.Count > 0)
            {
                int trackId = trackToBeLoaded.Dequeue();
                mediaDb.SearchTrackRecords(trackId, true);
            }
            else
            {
                // all tracks loaded
                labelStatusInfo.Text = string.Format("Complete. {0} tracks found.", mediaDb.Tracks.Count);
                SelectTrack(-1);
            }
        } 

        #endregion
     

        #region Select track methods and handlers

        private void comboBoxTracks_SelectedIndexChanged(object sender, EventArgs e)
        {
            SelectTrack(comboBoxTracks.SelectedIndex);
        }

        private void SelectTrack(int selectedTrackId)
        {
            buttonExport.Enabled = selectedTrackId >= 0;
            textBoxOutPath.Enabled = selectedTrackId >= 0;
            buttonOutFile.Enabled = selectedTrackId >= 0;
            comboBoxTracks.Enabled = State == ConnectionState.Connected;

            if (selectedTrackId >= 0)
            {
                if (selectedTrackId < mediaDb.Tracks.Count)
                {
                    exportTrack = mediaDb.Tracks[selectedTrackId];
                }
                else
                {
                    Trace.TraceWarning("Track not found in media database");
                    SelectTrack(-1);
                    return;
                }

                // update file name, but keep folder and extension
                string path = textBoxOutPath.Text.Contains("\\")
                                  ? Path.GetDirectoryName(textBoxOutPath.Text)
                                  : Directory.GetCurrentDirectory();

                string ext = Path.GetExtension(textBoxOutPath.Text);

                // set defualt format if not specified
                if (string.IsNullOrEmpty(ext))
                    ext = ".asf";

                string fileName = string.Format("{0}-Track{1}{2}", textBoxDeviceIP.Text, exportTrack.TrackID, ext);

                textBoxOutPath.Text = Path.Combine(path, fileName);


                // get track time bounds 
                // Note: In current version Track.StartTime and Track.EndTime throws NotImlementedException
                //      So, we gets time bounds from first and last events

                var trackRecords = mediaDb.Records.Where(x => x.TrackID == exportTrack.TrackID);
 
                if (trackRecords.Count() == 0)
                {
                    // disable export
                    SelectTrack(-1);
                    labelTrackInfo.Text = "Error: Unable to determinate track time bounds.";
                    return;
                }
                else
                {
                    exportTimeStart = trackRecords.Min(x => x.StartTime.UTC);
                    exportTimeEnd = trackRecords.Max(x => x.EndTime.UTC);

                    string info = string.Format("Track time bounds: {0} - {1}", exportTimeStart, exportTimeEnd);
                    labelTrackInfo.Text = info;
                }
            }
            else
            {
                // reset info
                textBoxOutPath.Text = null;
                labelTrackInfo.Text = mediaDb.Tracks.Count == 0 ? "Tracks are not loaded" : "Track is not selected";
                labelFormatInfo.Text = null;
                labelExportInfo.Text = null;
            }
        }

        #endregion

        #region Select output file and format
        private void buttonOutDir_Click(object sender, EventArgs e)
        {
            string dir = Path.GetDirectoryName(textBoxOutPath.Text);

            if (string.IsNullOrEmpty(dir))
                dir = Directory.GetCurrentDirectory();

            // set init dir and filename
            saveFileDialog.InitialDirectory = dir;
            saveFileDialog.FileName = Path.GetFileName(textBoxOutPath.Text);

            if (saveFileDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                textBoxOutPath.Text = saveFileDialog.FileName;
            }
        }

        private void textBoxOutPath_TextChanged(object sender, EventArgs e)
        {
            try
            {
                exportPath = Path.GetFullPath(textBoxOutPath.Text);
                string ext = Path.GetExtension(textBoxOutPath.Text);

                // ger format from selected extension
                exportFormat = string.Equals(ext, ".asf", StringComparison.CurrentCultureIgnoreCase)
                                        ? MediaFileFormatEnum.mffASF
                                        : MediaFileFormatEnum.mffMPEGActiveX;

                // show output format info
                string formatShortName = exportFormat.ToString().Replace("mff", "");
                labelFormatInfo.Text = string.Format("Export in {0} format", formatShortName);
            }
            catch (Exception ex)
            {
                exportPath = null;
                labelFormatInfo.Text = null;
            }

            buttonExport.Enabled = exportPath != null;
        }

        #endregion

        #region Export method and handlers
        private void buttonExport_Click(object sender, EventArgs e)
        {
            if (writer != null)
            {
                writer.StopRecording();
                writer = null;

                buttonExport.Text = "Start";
                return;
            }

            buttonExport.Text = "Stop";
            
            labelExportInfo.Text = "Start Export track #" + exportTrack.TrackID + "...";

            Export(exportTrack, exportPath, exportFormat, exportTimeStart, exportTimeEnd);
        }

        private void Export(Track track, string path, MediaFileFormatEnum format, DateTime firstEventStart, DateTime lastEventEnd)
        {
            Trace.TraceInformation("Export track #{0} to file {1}", track.TrackID, path);
            Trace.TraceInformation("Set export file format to {0}", format);
            Trace.TraceInformation("Set export time bounds {0} - {1}", firstEventStart, lastEventEnd);

            playback = new PlaybackController();

            // init writer
            writer = new MediaFileWriter();
            writer.FileFormat = format;
            writer.FileSizeLimitKB = 0; // no limit
            writer.MaximumNumberOfFiles = 0;
            writer.RecordingStartTime = new Time64() { UTC = firstEventStart };
            writer.RecordingEndTime = new Time64() { UTC = lastEventEnd };

            // set handlers
            writer.Progress += writer_Progress;
            writer.RecordingStopped += writer_RecordingStopped;
            writer.NewFileCreated += writer_NewFileCreated;

            var session = track.GetMediaSession(playback);

            // set stream
            writer.AddStream(session.GetVideoStream(),
                             MediaTypeEnum.mteVideo,
                             track.TrackID,
                             "Track " + track.TrackID,
                             track.SourceURL,
                             track.SourceID);

            // start export
            writer.StartRecording(textBoxOutPath.Text, String.Empty);

            SetConnectionState(ConnectionState.Export, "Export in progress");

            textBoxOutPath.Enabled = false;
            buttonOutFile.Enabled = false;
            comboBoxTracks.Enabled = false;
            textBoxDeviceIP.Enabled = false;
        }

        private void writer_NewFileCreated(string Filename)
        {
            Trace.TraceInformation("Export file created {0}", Filename);
        }

        private void writer_RecordingStopped(RecordingStoppedEnum reason)
        {
            Trace.TraceInformation("Export recording stoped with reason {0}", reason);

            if (reason == RecordingStoppedEnum.rseEndTimeReached)
                labelExportInfo.Text = "Completed";
            else if (reason == RecordingStoppedEnum.rseClientStoppedRecording)
                labelExportInfo.Text = "Aborted";
            else if (reason == RecordingStoppedEnum.rseEndTimeReached)
                labelExportInfo.Text = "Finished with errors";
            else
                labelExportInfo.Text = "Failed due to " + reason;

            writer = null;
            buttonExport.Text = "Start";

            SetConnectionState(ConnectionState.Connected, "");

            textBoxOutPath.Enabled = true;
            buttonOutFile.Enabled = true;
            comboBoxTracks.Enabled = true;
            textBoxDeviceIP.Enabled = true;
        }

        private void writer_Progress(int TotalEstSec, int ElapsedSec, int KBytesWritten, Time64 pCurrentRecTime, int ErrorCount)
        {
            int progress = ElapsedSec * 100 / TotalEstSec;
            progressBarLoading.Value = progress;

            string errorsInfo = ErrorCount == 0 ? String.Empty : ErrorCount.ToString("{0} errors.");
            string remainingTimeInfo = new TimeSpan(0, 0, TotalEstSec - ElapsedSec).ToString();

            string info = string.Format("Export in progress {0}%... {1} {2} Kb written. Remaining time : {3}",
                progress,
                errorsInfo,
                KBytesWritten,
                remainingTimeInfo);

            labelExportInfo.Text = info;
            Trace.TraceInformation(info);
        }

        #endregion

        #region UI aux methods

        private void Clear(bool clearLog)
        {
            if (clearLog)
                textBoxLog.Clear();

            comboBoxTracks.Items.Clear();

            trackToBeLoaded.Clear();
            tracksCompletelyLoaded = 0;

            progressBarLoading.Value = 0;

            // deselect track item
            SelectTrack(-1);
        }

        private void OnTraceListenerLogWrite(object sender, LogEventArgs e)
        {
            var str = e.Message + "\r\n";
            Action updateAction = (() => textBoxLog.AppendText(str));

            SafeInvoke(textBoxLog, updateAction);
        }

        /// <summary>
        /// Safe invoke method for cpecified control
        /// </summary>
        /// <param name="uiElement">UI Control</param>
        /// <param name="updateAction">Action to be executed</param>
        private static void SafeInvoke(Control uiElement, Action updateAction)
        {
            if (uiElement == null)
            {
                throw new ArgumentNullException("uiElement");
            }

            if (uiElement.InvokeRequired)
            {
                uiElement.BeginInvoke((Action)(() => SafeInvoke(uiElement, updateAction)));
            }
            else
            {
                updateAction();
            }
        }
        #endregion
    }
}
